#!/usr/bin/env python

# Transpose a MUP score. Not perfect !!

import sys, os, getopt

version = 0.1

steps = 0
startline = 0
endline = 999999999

#### BUGS ####

# - if you are in the key of F and transpose an unmarked 'b' up 2 steps
#   you'll get a 'd&' instead of a 'c'. Problem is that the program doesn't
#   understand keysigs ... in the case of 'F' the 'b' is flat but the program
#   didn't know. This could be fixed by having some keysig tables, and doing more
#   lookups.



##############################
# system helpers

def error( s = ''):
    print "Mup-trans ERROR"
    print s
    print
    sys.exit(1)

def usage():
    print "Mup-trans, (c) Bob van der Poel"
    print "Transpose a MUP score."
    print "Usage: mup-trans <opts> filename <opts>"
    print " -t -- number of steps up/down to tranpose."
    print " -s -- line number of first line transpose."
    print " -e -- line number of last line to transpose."
    print " -v -- prints version."
    print
    sys.exit(1)

def getint(a, msg):
    try: return int(a)
    except: error("Expecting integer arg for '%', not '%s'." % msg)


#################################
# Transpose funcs

upchordtable = {'A':'B&', 'B':'C', 'C':'D&', 'D':'E&', 'E':'F', 'F':'G&', 'G':'A&', 
                'A#':'B', 'B#':'C#', 'C#':'D', 'D#':'E', 'E#':'F#', 'F#':'G', 'G#':'A',
                'A&':'A', 'B&':'B', 'C&':'C', 'D&':'D', 'E&':'E', 'F&':'F', 'G&':'G' }

downchordtable = {'A':'A&', 'B':'B&', 'C':'B', 'D':'D&', 'E':'E&',  'F':'E', 'G':'G&', 
                  'A#':'A', 'B#':'B', 'C#':'C', 'D#':'D', 'E#':'E', 'F#':'F', 'G#':'G',
                  'A&':'G', 'B&':'A', 'C&':'B&', 'D&':'C', 'E&':'D', 'F&':'E', 'G&':'F' } 

def exg(l, p, new, oldsize):
    return l[:p] + new + l[p+oldsize:]


def chord(l):

    if steps > 0:
        chordtable = upchordtable
    else:
        chordtable = downchordtable

    xs = abs(steps)
    end = len(l)-1

    for x in range(xs):
        p = 0
        inchord = 0
        while p < end:
            if l[p] == '"':
                if inchord:
                    inchord = 0
                else:
                    inchord = 1
                p+=1
                continue

            if inchord:
                r = l[p:p+2]
                if r in chordtable:
                    new = chordtable[r]
                    l = exg(l, p, new, 2 )
                    p += 2
                    continue

                r = l[p:p+1]
                if r in chordtable:
                    new = chordtable[r]
                    l = exg(l, p, new, 1 )
                    p += 1
                    continue
                
            p += 1

    return l
    

def keysig(l):

    if steps > 0:
        chordtable = upchordtable
    else:
        chordtable = downchordtable

    xs = abs(steps)

    try:
        p = l.index('=')
    except:
        return "### FIXME " + l

    while l[p] in ' =':
        p+=1

    for x in range(xs):    
        k = l[p:p+2].upper()
        if k in chordtable:
            l = exg(l, p, chordtable[k].lower(), 2)

        else:
            k = l[p].upper()
            if l[p] in chordtable:
                l = exg(l, p, chordtable[k].lower(), 1)

    return l

upnotetable = {'a':'b&', 'b':'c+', 'c':'d&', 'd':'e&', 'e':'f', 'f':'g&', 'g':'a&', 
                'a#':'b', 'b#':'c#+', 'c#':'d', 'd#':'e', 'e#':'f#', 'f#':'g', 'g#':'a',
                'a&':'a', 'b&':'b', 'c&':'c', 'd&':'d', 'e&':'e', 'f&':'f', 'g&':'g',
                '(#)':'(n)', ('n'):'(#)', '(&)':'(n)' }

downnotetable = {'a':'a&', 'b':'b&', 'c':'b-', 'd':'d&', 'e':'e&',  'f':'e', 'g':'g&', 
                 'a#':'a', 'b#':'b', 'c#':'c', 'd#':'d', 'e#':'e', 'f#':'f', 'g#':'g',
                 'a&':'g', 'b&':'a', 'c&':'b&-', 'd&':'c', 'e&':'d', 'f&':'e', 'g&':'f',
                 '(#)':'(n)', ('n'):'(&)', '(&)':'(n)' }        


def notes(l):
    
    if steps > 0:
        notetable = upnotetable
    else:
        notetable = downnotetable

    xs = abs(steps)

    for x in range(xs):
        p = 0
        end = len(l)
        while p < end:
            if l[p] == '[':  # Skip over [with <stuff> ]
                while p < end and l[p] != ']':
                    p += 1
                continue

            r = l[p:p+3]    # check 3 char items like '(#)'
            if r in notetable:
                new = notetable[r]
                l = exg(l, p, new, 3 )
                p += 3
                end = len(l)
                continue
                
            r = l[p:p+2]    # 2 char items like 'a&'
            if r in notetable:
                new = notetable[r]
                l = exg(l, p, new, 2 )
                p += 2
                end = len(l)
                continue

            r = l[p:p+1]    # 1 char items like 'e'
            if r in notetable:
                new = notetable[r]
                l = exg(l, p, new, 1 )
                p += 1
                end = len(l)
                continue
                
            p += 1
    l = l.replace('+-', '')
    l = l.replace('-+', '')
    return l


#####################################
# mainline


try:
    opts, args = getopt.gnu_getopt(sys.argv[1:],
            "vt:s:e:", [] )
except getopt.GetoptError:
    usage()

for o,a in opts:
    if o == '-t':
        steps = getint(a, 'steps')

    elif o == '-s':
        startline = getint(a, 'startline')

    elif o == '-e':
        endline = getint(a, 'endline')

    elif o == '-v':
        print version
        sys.exit(0)

    else:
        print "Unknown option, '%s'" % o
        usage()

if len(args) != 1:
    print "No filename given."
    usage()

filename = args[0]

try:
    inpath = open(filename, 'r')
except:
    error("Can't access file '%s'." % filename)

lnum =0

for l in inpath:

    l = l.rstrip()   # dump trailing NL from each line
    
    if lnum>=startline and lnum <= endline:   # do in range

        l = l.split('//', 1)     # strip/save comments
        if len(l)>1:
            comment=l[1]
        else:
            comment=None
        l=l[0]

        if l.startswith("C ") or l.startswith("chord "):
            l = chord(l)
                                          
        elif l.strip().startswith("key"):
            l = keysig(l)

        else:
            if l.find(':') > -1:
                l1,l2 = l.split(':')
                l = l1 + ":" + notes(l2)

        if comment:             # restore comment
            l += '//' + comment

    print l

    lnum += 1

    
